#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define pow(a,b)	(exp((b)*log(a)))

#include <windows.h>
#include "queue_template.h"


#define DETAIL	0

#define TOL	(1.0e-10)

#define M	7	//	number of rows
#define N	5	//	number of columns



#define P	4.0			//must have P >= 2.0 for this program to work correctly
#define Q	4.0			//must have Q >= 1.0 for this program to work correctly

double mat[M][N]={
{0.62688629875545775929, 0.0, 0.0, 0.0, 0.0},
{0.55117543446238233305, 0.40317274455075456055, 0.0, 0.0, 0.0},
{-0.59022612533822156735, 0.39248860786580614888, 0.18358329898434102747, 0.0, 0.0},
{0.0, -0.71233971732316618559, 0.30290091733745758709, 0.12695449210375910942, 0.0},
{0.0, 0.0, -0.76308037230788877211, 0.29075358717135370929, 0.079701037732094712077},
{0.0, 0.0, 0.0, -0.82283134832528448594, 0.20504870863967714789},
{0.0, 0.0, 0.0, 0.0, -0.90207952540163340728}
		};


double conversion_factor;//	used to convert from halfside of cube to the radius of the patch


//	This is the norm on the output side. Here it computes the L^Q norm but it could be any norm.
double normeval(double matrix[M][N], double z[N])
{
int j,k;
double s,t;

t=0.0;
for(j=0;j<M;j++)
	{
	s=0.0;
	for(k=0;k<N;k++) s+=matrix[j][k]*z[k];
	t+=pow(fabs(s),Q);
	}
return(pow(t,1.0/Q));
}



typedef struct{

int code;//	in range 0 <= code < N, tells which variable is fixed at 1.0

double centre[N];

double halfside;

}surface_subcube;


typedef struct{

double centre[N];

double radius;// some upper bound for max \|x-y\|_p where x = the projection of the centre of the surface subcube
				//	 and as y runs over the points of the patch

}patch;


void print_cube(surface_subcube &cube)
{
int j;

for(j=0;j<N;j++)
	{
	if (cube.code == j) printf("{%f}, ",cube.centre[j]);
	else printf("[%f,%f], ",cube.centre[j]-cube.halfside,cube.centre[j]+cube.halfside);
	}

printf("halfside =%f\n",cube.halfside);
}

void print_patch(patch &cap)
{
int j;

for(j=0;j<N;j++) printf("%f, ",cap.centre[j]);

printf("radius =%f\n",cap.radius);
}

void cube_to_patch(surface_subcube &cube, patch *cap)
{
double r;
int j;

r=0.0;
for(j=0;j<N;j++) r+=pow(fabs(cube.centre[j]),P);
r=pow(r,1.0/P);
for(j=0;j<N;j++) cap->centre[j]=cube.centre[j]/r;

cap->radius=conversion_factor*cube.halfside;
}



void initialize_queue(Queue<surface_subcube> &QueueOfCubes)
{
long j;
surface_subcube newcube;

for(j=0;j<N;j++) newcube.centre[j]=0.0;
newcube.halfside=1.0;
for(j=0;j<N;j++)
	{
	newcube.centre[j]=1.0;
	newcube.code=j;
	QueueOfCubes.Enqueue(newcube);
	newcube.centre[j]=0.0;
	}
}




int cube_can_be_eliminated(patch &cap, double &maxseen, double &value)
{
double ratio;

ratio=1.0-0.5*(P-1.0)*cap.radius*cap.radius;

if (ratio<=0.0) return(0);


return((value < maxseen*ratio*(1.0+TOL)) ? 1 : 0);
}


int is_viable(patch &cap)
{
double ratio;
ratio=1.0-0.5*(P-1.0)*cap.radius*cap.radius;

return ((ratio>0.0) ? 1 : 0);
}




double operator_norm(double matrix[M][N], double *maxpt, long *nevals)
{
double maxpoint[N];
long numevals;
long j,k,kk,max;
surface_subcube curcube, newcube;
patch cap;
Queue<surface_subcube> QueueOfCubes;
double maxseen,t;

maxseen=0.0;
numevals=0;
max=0;
initialize_queue(QueueOfCubes);

while(1)
	{
	if (QueueOfCubes.Dequeue(curcube)) break;
	cube_to_patch(curcube,&cap);
#if DETAIL
	printf("\n");
	print_cube(curcube);
	print_patch(cap);
#endif
	if (is_viable(cap))//	check to see if it's even worthwhile to figure normeval(matrix, cap.centre)
		{
		numevals++;
		t=normeval(matrix, cap.centre);
	#if DETAIL
		printf("t = %f, max= %f   ",t,maxseen);
	#endif
		if (t>maxseen)
			{
			maxseen=t;
			for(j=0;j<N;j++) maxpoint[j]=cap.centre[j];
			}

		if (cube_can_be_eliminated(cap,maxseen,t))
			{
	#if DETAIL
			printf("eliminated\n");
	#endif
			continue;
			}
		}
#if DETAIL
	printf("subdividing\n");
#endif
	for(k=0;k<(1L<<(N-1));k++)
		{
		newcube.halfside=curcube.halfside*0.5;
		newcube.code=curcube.code;
		kk=k;
		for(j=0;j<N;j++)
			{
			if ((curcube.code%N)==j)
				{
				newcube.centre[j]=curcube.centre[j];
				continue;
				}
			if (kk%2) newcube.centre[j]=curcube.centre[j]+newcube.halfside;
			else      newcube.centre[j]=curcube.centre[j]-newcube.halfside;
			kk/=2;
			}

		QueueOfCubes.Enqueue(newcube);
		}
	}

if (nevals) *nevals=numevals;
if (maxpt) for(j=0;j<N;j++) maxpt[j]=maxpoint[j];

return(maxseen*(1.0+TOL));//	that's an upper bound for the operator norm
}



extern int main(void)
{
long j,ne;
double maxpt[N];
double norm;

conversion_factor=2.0*pow((double)(N-1),1.0/P);

norm = operator_norm(mat, &maxpt[0] , &ne);

printf("\n\nnorm = %14.12f, numevals=%ld\n\n",norm,ne);

for(j=0;j<N;j++) printf("pt[%d]=%14.12f\n",j,maxpt[j]);

return(0);
}