

#define ISIZE	100
#define QUEUE_EMPTY	1

/*
The following facts are assumed about the class T
used in Queue. Instances of the class T are
referred to as objects.


1)	There is an (overloaded) copy assignment operator

		T const &operator=(T const &other);


2)	There is a destructor

		~T(void).


*/




template <class T>
class Queue{

public:

//	constructor
Queue(void);

//	destructor
~Queue(void);

void Enqueue(T &object);

int Dequeue(T &object);

inline long NumberInQueue(void){return(number_queued);}

private:

HGLOBAL table;
long table_size;
long first;
long last;
long number_queued;
};


template <class T>
Queue<T>::Queue(void)
{

table=GlobalAlloc(GMEM_MOVEABLE,ISIZE*sizeof(T));
if (!table) exit(0);

first=last=number_queued=0L;
table_size=ISIZE;
}

template <class T>
Queue<T>::~Queue(void)
{
T *p;

p=(T*)GlobalLock(table);

//	delete anything left in the table
while(first!=last)
	{
	p[first].~T();
	first++;
	if (first==table_size) first=0;
	}

//	delete the table itself
GlobalFree(table);
}


template <class T>
void Queue<T>::Enqueue(T &object)
{
T *p,*q;
HGLOBAL h;
long j;

p=(T*)GlobalLock(table);
if (number_queued>=(table_size-1))
	{
	//	the table is (nearly) full so make a new one
	h=GlobalAlloc(GMEM_MOVEABLE,2*table_size*sizeof(T));
	if (!h) exit(0);
	//	copy the data
	q=(T*)GlobalLock(h);
	j=0;
	while(first!=last)
		{
		q[j]=p[first];
		p[first].~T();
		first++;
		j++;
		if (first==table_size) first=0;
		}
	//	delete the old table and replace with new
	GlobalFree(table);
	table=h;
	table_size*=2;
	first=0;
	last=number_queued;
	}

p[last++]=object;
number_queued++;
GlobalUnlock(table);
if (last==table_size)
	{
	//	if upsizing is a good thing to do, it's easy now-- do it!
	if (table_size<2*number_queued)
		{
		table_size*=3;
		table_size/=2;
		table=GlobalReAlloc(table,table_size*sizeof(T),GMEM_MOVEABLE);
		}
	else last=0L;
	}
return;
}


template <class T>
int Queue<T>::Dequeue(T &object)
{
T *p,*q;
long j;
HGLOBAL h;

if (!number_queued) return(QUEUE_EMPTY);
p=(T*)GlobalLock(table);

object=p[first++];
number_queued--;

if (first==table_size) first=0;

if (table_size>ISIZE)
	{
	//	if downsizing is easy  and desirable-- do it!
	if ((first<last)&&((2*last)<table_size))
		{
		table_size*=2;
		table_size/=3;
		table=GlobalReAlloc(table,table_size*sizeof(T),GMEM_MOVEABLE);
		//printf("table downsized, size=%ld\n",table_size);
		}
	//	if downsizing is overdue -- do it!
	else if ((4*number_queued)<table_size)
		{
		h=GlobalAlloc(GMEM_MOVEABLE,(table_size/2)*sizeof(T));
		if (!h) goto done;
		//	copy the data
		q=(T*)GlobalLock(h);
		j=0;
		while(first!=last)
			{
			q[j]=p[first];
			p[first].~T();
			first++;
			if (first==table_size) first=0;
			j++;
			}
		//	delete the old table and replace with new
		GlobalFree(table);
		table=h;
		first=0;
		last=number_queued;
		table_size/=2;
		}
	}
done:
GlobalUnlock(table);
return(0);
}



